import { ClientBase } from 'pg';
import { TransactionalLogger } from '../common/logger';
import { InboxMessage, OutboxMessage } from '../common/message';
import { InboxConfig } from './inbox-listener';
/**
 * Initialize the inbox message storage to store incoming messages in the inbox table.
 * @param config The configuration object that defines the values on how to connect to the database and general settings.
 * @param logger A logger instance for logging trace up to error logs
 * @returns The function to store the inbox message data to the database and the shutdown action.
 */
export declare const initializeInboxMessageStorage: (config: Pick<InboxConfig, 'pgConfig' | 'settings'>, logger: TransactionalLogger) => [storeInboxMessage: (message: OutboxMessage) => Promise<void>, shutdown: () => Promise<void>];
/**
 * This function increases the "started_attempts" for the inbox message by one
 * in the inbox table. This number can then be compared to the
 * "finished_attempts" number which is only increased when a message processing
 * exception was correctly handled or an error was caught. A difference between
 * the two can only happen if the service crashes after increasing the
 * "started_attempts" but before successfully marking the message as done
 * (success case) or catching an error (error case). If the "started_attempts"
 * and the "finished_attempts" field differ by more than one, the chances are
 * high that this message is causing a service crash.
 * It sets the started_attempts, finished_attempts and processed_at
 * values on the inbox message.
 * For additional safety it makes sure, that the inbox item was not and is not
 * currently being worked on.
 * @param message The inbox message for which to acquire a lock and increment the started_attempts
 * @param client The database client. Must be part of a transaction that runs before the message handling transaction.
 * @param config The configuration settings that defines inbox database schema.
 * @returns 'INBOX_MESSAGE_NOT_FOUND' if the message was not found, 'ALREADY_PROCESSED' if it was processed, and otherwise assigns the properties to the message and returns true.
 */
export declare const startedAttemptsIncrement: (message: InboxMessage, client: ClientBase, { settings }: Pick<InboxConfig, 'settings'>) => Promise<true | 'INBOX_MESSAGE_NOT_FOUND' | 'ALREADY_PROCESSED'>;
/**
 * This function makes sure the inbox item was not and is not currently being
 * worked on and acquires a lock to prevent other processes to work with this
 * inbox message. As the inbox listener does not run in the same transaction as
 * the message handler code there is a small chance that the handler code
 * succeeds but the WAL inbox message was not acknowledged. This takes care of
 * such cases by ensuring that the message was not yet processed.
 * It sets the started_attempts, finished_attempts and processed_at
 * values (again) on the inbox message to be sure no other process altered them.
 * @param message The inbox message for which to acquire a lock and fill the started_attempts, finished_attempts and processed_at values
 * @param client The database client. Must be part of the transaction where the message handling changes are later done.
 * @param config The configuration settings that defines inbox database schema.
 * @returns 'INBOX_MESSAGE_NOT_FOUND' if the message was not found, 'ALREADY_PROCESSED' if it was processed, and otherwise assigns the properties to the message and returns true.
 */
export declare const initiateInboxMessageProcessing: (message: InboxMessage, client: ClientBase, { settings }: Pick<InboxConfig, 'settings'>) => Promise<true | 'INBOX_MESSAGE_NOT_FOUND' | 'ALREADY_PROCESSED'>;
/**
 * Marks the inbox message as handled by setting the processed_at date to the current date and time.
 * It also increases the finished_attempts count by one.
 * @param message The inbox message to acknowledge.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 */
export declare const markInboxMessageCompleted: ({ id }: InboxMessage, client: ClientBase, { settings }: Pick<InboxConfig, 'settings'>) => Promise<void>;
/**
 * Does increase the finished_attempts count by one.
 * @param message The inbox message to NOT acknowledge.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 */
export declare const increaseInboxMessageFinishedAttempts: ({ id }: InboxMessage, client: ClientBase, { settings }: Pick<InboxConfig, 'settings'>) => Promise<void>;
//# sourceMappingURL=inbox-message-storage.d.ts.map